package OAIServer;
#
# Minimal OAI protocol implementation for data-provider
# Simeon Warner - May2001...
#
# [CVS: $Id: OAIServer.pm,v 1.3 2001/06/05 20:07:02 simeon Exp $]
use strict;

use XML::Writer;
use Database qw(&recordExists &recordIsDeleted &getIdsByDate &getDatestamp 
                &getMetadata &getMetadataFormats &getMetadataFormatInfo);

use vars qw(@ISA @EXPORT_OK);
use Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(&OAICheckRequest &OAISatisfyRequest);

# Configuration variables
#
my $OAI_REPOSITORY_NAME="Dummy Repository";
my $OAI_BASE_URL="http://localhost/oai1";     
my $OAI_VERSION = "1.0";
my $OAI_ADMIN_EMAIL="root\@localhost";

# Grammar for requests, for arguments: 0->optional, 1->required, 2->exclusive
#
my $SUBROUTINE=0; # array position in grammar for subroutine
my $ARGS=1;       # array position in grammar for args
my %grammar = (
  'Identify' => 
     [ \&identify, {} ],
  'ListSets' => 
     [ \&listSets, {} ],
  'ListMetadataFormats' =>
     [ \&listMetadataFormats, {
       'identifier' => 0 } ],
  'GetRecord' =>
     [ \&getRecord, { 
       'identifier' => 1, 
       'metadataPrefix' => 1 } ],
  'ListRecords' =>
     [ \&listRecords, { 
       'set' => 0, 
       'from' => 0, 
       'until' => 0,
       'metadataPrefix' => 1,
       'resumptionToken' => 2 } ],
  'ListIdentifiers' =>
     [ \&listIdentifiers, {
       'set' => 0, 
       'from' => 0, 
       'until' => 0,
       'resumptionToken' => 2 } ],
);


# Check the validity of the OAI request
#
# Returns undef if okay, string describing error otherwise
#
sub OAICheckRequest {
  my $args = shift;
  my $error = undef;
  #
  # First check to see what the verb is, all requests must have a verb
  #
  my $verb = $args->{verb};
  return("No verb specified!") unless (defined $verb);
  #
  # Get grammar for the particular verb
  #
  my $grammarPtr = $grammar{$verb};
  return("Bad verb.\nVerb '$verb' not implemented.") unless (defined $grammarPtr);
  #
  # Check arguments supplied against grammar.
  #
  my $exclusive=undef;
  foreach my $key (keys %{$args}) {
    next if ($key eq 'verb');
    unless (defined $grammarPtr->[$ARGS]->{$key}) {
      $error .= "Argument '$key' not known.\n";
    } else {
      if ($grammarPtr->[$ARGS]->{$key}==2) {
        if (defined $exclusive) {
          $error .= "Arguments '$key' and '$exclusive' are exclusive!\n";
        } else {
          $exclusive=$key;
        }
      }
    }
  }
  #
  # If any exclusive arguments have been given then we must have no
  # other arguments. Otherwise check that all required arguments have 
  # been provided
  #
  if (defined($exclusive)) {
    if (scalar(keys %{$args})!=2) {
      $error .= "Argument '$exclusive' is exclusive, no others allowed.\n";  
    }
  } else {
    foreach my $key (keys %{$grammarPtr->[$ARGS]}) {
      if (not defined($args->{$key}) and $grammarPtr->[$ARGS]->{$key}==1) {
        $error .= "Argument '$key' required but not supplied.\n";
      }
    }
  }
  #
  # For the List requests we need to do additional checks to see that 
  # any supplied resumptionToken is valid. In this implementation we
  # don't issue and hence don't accept resumptionTokens.
  #
  if ($verb=~/^List/) {
    $error .= "resumptionToken not supported.\n" if (defined($args->{resumptionToken}));
  }
  #
  return("Bad arguments.\n".$error) if ($error);
  return(undef);
}


# Satisfy the pre-checked request, response is XML output to STDOUT
#
sub OAISatisfyRequest {
  my $args = shift;
  #
  my $verb=$args->{verb};
  my $writer = new XML::Writer(DATA_MODE=>1, DATA_INDENT=>1);
  $writer->xmlDecl("UTF-8");
  $writer->startTag($verb,
    'xmlns'=>"http://www.openarchives.org/OAI/OAI_$verb",
    'xsi:schemaLocation'=>"http://www.openarchives.org/OAI/1.0/OAI_$verb http://www.openarchives.org/OAI/1.0/OAI_$verb.xsd");
  # Elements common to all replies
  $writer->dataElement("responseDate",&dateTimeNow());
  $writer->dataElement("requestURL",&requestURL($verb,$args));
  # Bulk of reply (the request), subroutine pointer from grammar
  my ($error) = &{$grammar{$verb}->[$SUBROUTINE]}($writer,$args);
  # Finish reply
  $writer->endTag($verb);
  $writer->end();
}


# Generates current time as an XML Schema timeInstant
# http://www.w3.org/TR/2000/CR-xmlschema-2-20001024/#timeInstant 
# Example: 2000-10-01T19:20:30+01:00
#
sub dateTimeNow {
  my $t=time;
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($t);
  my ($gmtmin,$gmthour)=(gmtime($t))[1,2];
  # Find offset from difference between GMT and local timezone;
  my $ohour=($hour-$gmthour);
  my $omin=($min-$gmtmin);
  if ($ohour<-11) { $ohour+=24; }
  if ($ohour>12) { $ohour-=24; }
  if ($ohour<0 and $omin>0) { $omin=60-$omin; $ohour+=1; } 
  if ($ohour>0 and $omin<0) { $omin=60+$omin; $ohour-=1; } 
  #
  $year+=1900; #perl counts from 1900
  my $sign='+';
  if ($ohour<0) { $ohour=-$ohour; $omin=-1*$omin; $sign='-'; }
  return(sprintf("%04d-%02d-%02dT%02d:%02d:%02d%s%02d:%02d",
                  $year,$mon,$mday,$hour,$min,$sec,$sign,$ohour,$omin));
}

# Builds equivalent GET URL for request
#
sub requestURL {
  my ($verb,$args) = @_;
  my $url = $OAI_BASE_URL.'?verb='.$verb;
  foreach my $key (keys %$args) {
    $url .= "&$key=".&escapeURLChars($args->{$key});
  }
  return($url);
}

# Hex encode special characters in URLs
#
sub escapeURLChars {
  my $str = shift;
  $str =~ s/([ +\/\?#=&:;])/&hexEncode($1)/ge;
  return($str);
}
sub hexEncode {
  my $xx=unpack('H2',shift);
  $xx=~tr/abcdef/ABCDEF/; #just to make the URL easier to read 
  return("%$xx");
}


##############################################################################
#
# Code to implement the OAI verbs
#

# Identify (only two element are compulsory) 
#
sub identify {
  my $writer = shift;
  $writer->dataElement('repositoryName',$OAI_REPOSITORY_NAME);
  $writer->dataElement('baseURL',$OAI_BASE_URL);
  $writer->dataElement('protocolVersion',$OAI_VERSION);
  $writer->dataElement('adminEmail',$OAI_ADMIN_EMAIL);
}


# ListSets (no sets implemented so empty return)
#
sub listSets {
  #nothing
}


# ListIdentifiers and ListRecords
#
# Essentially a search by date and set, with the option of
# returning just identifiers or identifiers and metadata in a 
# particular metadataFormat.
#
# The variable $MAX_RECORDS is passed to Updates.pm to limit
# the number of records returned. If the limit is reached then 
# the $limited value returned from Updates.pm is simply the date
# of the last record returned and is thus the date required as
# the $recordFrom date to resume the search (since the new search
# starts from the day after). The resumptionToken returned in the
# XML reply is simply a concatenation of 
# recordFrom_recordUntil_set_metadataFormat with '_' to join.
#
# If metadataPrefix is not specified then we are doing listIdentifiers
#
# The only purpose of the two stub routines is to allow the
# resumptionToken to be checked to see if it comes from the
# correct request type.

sub listIdentifiers {
  return(&listEither(@_,0));
}

sub listRecords {
  return(&listEither(@_,1));
} 

sub listEither {
  my ($writer, $args, $meta) = @_;
  #
  # Empty reply if set specified (sets not implemented)
  return(undef) if (defined($args->{set}));
  #
  foreach my $id (&getIdsByDate($args->{from},$args->{until})) {
    if ($meta) {
      &disseminate($writer,$id,$args->{metadataPrefix});
    } else {
      my %atts=();
      $atts{status}='deleted' if (&recordIsDeleted($id)); 
      $writer->dataElement('identifier',$id,%atts);
    } 
  }
  return(undef);
}


# GetRecord
#
# Writes metadata for that record in the specified metadataPrefix.
# 
sub getRecord {
  my $writer = shift;
  my $args = shift;
  #
  # Simply return (blank reply) if record doesn't exist
  return(undef) unless (&recordExists($args->{identifier}));
  #
  # Else write record with header and possibly metadata
  return(&disseminate($writer,$args->{identifier},$args->{metadataPrefix}));
}



# Writes XML record for $id in $metadataPrefix. It is assumed that $id 
# and $metadataPrefix have already by validated.
#
# Only the <header>...</header> is returned if $metadataPrefix is
# undef. Otherwise also <metadata>...</metadata> in $metadataPrefix
# format. The <about> block is not implemented.
#
sub disseminate {
  my $writer = shift;
  my $id = shift;
  my $metadataPrefix = shift; 
  #
  $writer->startTag('record');
  #
  # Make <header> block
  #
  my %atts=();
  $atts{status}='deleted' if (&recordIsDeleted($id));
  $writer->startTag('header',%atts);
  $writer->dataElement('identifier',$id);
  $writer->dataElement('datestamp',&getDatestamp($id));
  $writer->endTag('header');
  #
  # Add on <metadata> block if required
  #
  my $meta = &getMetadata($id,$metadataPrefix);
  if (defined($meta)) {
    $writer->startTag('metadata');
    my $atts=(&getMetadataFormatInfo($metadataPrefix))[1];
    $writer->startTag($metadataPrefix, %$atts);
    my $el;
    while ($el=shift(@$meta)) {
      $writer->dataElement($el,shift(@$meta));
    }
    $writer->endTag($metadataPrefix);
    $writer->endTag('metadata');
  }
  $writer->endTag('record');
  return(undef);
}


# ListMetadataFormats
#
# Return a list of metadata formats available from this repository
# or for the specified record.
#
sub listMetadataFormats {
  my $writer = shift;
  my $args = shift;
  my @mfs=&getMetadataFormats($args->{identifier});
  foreach my $mf (@mfs) {
    $writer->startTag('metadataFormat');
    $writer->dataElement('metadataPrefix', $mf);
    $writer->dataElement('schema', (&getMetadataFormatInfo($mf))[0]);
    $writer->endTag('metadataFormat');
  }
  return(undef);
}

1;
